/*
 * The following code has been adapted for use in the A1200 IDE MINIX driver.
 * The original code was written by Hale Landis.
 *
 * A. Eleuteri 12/2020
*/
 
/********************************************************************* */
/* MINIMUM ATA LOW LEVEL I/O DRIVER -- MINDRVR.C */
/* */
/* by Hale Landis (hlandis@ata-atapi.com) */
/* */
/* There is no copyright and there are no restrictions on the use */
/* of this ATA Low Level I/O Driver code.  It is distributed to */
/* help other programmers understand how the ATA device interface */
/* works and it is distributed without any warranty.  Use this */
/* code at your own risk. */
/* */
/* Minimum ATA Driver (MINDRVR) is a subset of ATADRVR. MINDRVR */
/* has a single header file and a single C file. MINDRVR can */
/* be used as the starting point for an ATADRVR for an embedded */
/* system. NOTE all the places in the MINDRVR.H and MINDRVR.C files */
/* where there is a comment containing the string "!!!". */
/* */
/* Use the header file mindrvr.h in any C files that call MINDRVR */
/* functions. */
/* */
/* This code is based on the ATA/ATAPI-4,-5 and -6 standards and */
/* on interviews with various ATA controller and drive designers. */
/* */
/* Note that MINDRVR does not support ATA CHS addressing. */
/* */
/* Most of the MINDRVR code is standard C code and should compile */
/* using any C compiler. It has been tested using Borland C/C++ 4.5. */
/* */
/* This C source file is the header file for the driver */
/* and is used in the MINDRVR.C files and must also be used */
/* by any program using the MINDRVR code/functions. */
/********************************************************************* */

#include "mindrvr.h"

/*************************************************************** */
/* */
/* !!! data that functions outside of MINDRVR must use */
/* */
/* Note that there is no actual "interrupt handler" provide in */
/* MINDRVR. The interrupt handler is usually a small function that */
/* is very system specific. However, MINDRVR expects that interrupt */
/* handler function to provide some status data at the time the */
/* interrupt handler is executed. */
/* */
/* In many systems, including PCI bus based systems, when an */
/* interrupt is received from an ATA controller, the interrupt */
/* handler must acknowledge the interrupt by reading both the */
/* ATA/ATAPI device Status register and the controller status */
/* register. This status must be stored here so that MINDRVR */
/* can use it. */
/* */
/*************************************************************** */

unsigned char int_ata_status;    /* ATA status read by interrupt handler */

unsigned char int_use_intr_flag = INT_DEFAULT_INTERRUPT_MODE;

struct REG_CMD_INFO reg_cmd_info;

int reg_config_info[2];

/* See Gayle specifications document p.7 
 * In A600/A1200 data port at offset 0x2000 is for 16 bit data transfers, while the
 * other one at offset 0x0000 is for 8-bit transfers.
 */
unsigned char * pio_reg_addrs[9] =
{
   PIO_BASE_ADDR1 + 0x2000, /* [0] CB_DATA */
   PIO_BASE_ADDR1 + 4*1,  /* [1] CB_FR & CB_ER */
   PIO_BASE_ADDR1 + 4*2,  /* [2] CB_SC */
   PIO_BASE_ADDR1 + 4*3,  /* [3] CB_SN */
   PIO_BASE_ADDR1 + 4*4,  /* [4] CB_CL */
   PIO_BASE_ADDR1 + 4*5,  /* [5] CB_CH */
   PIO_BASE_ADDR1 + 4*6,  /* [6] CB_DH */
   PIO_BASE_ADDR1 + 4*7,  /* [7] CB_CMD & CB_STAT */
   PIO_BASE_ADDR2 + 0   /* [8] CB_DC & CB_ASTAT */
} ;

unsigned char pio_xfer_width = PIO_DEFAULT_XFER_WIDTH;

/*************************************************************** */
/* */
/* functions internal and private to MINDRVR */
/* */
/*************************************************************** */

static void sub_setup_command(  );
static void sub_trace_command(  );
static int sub_select( );
static void sub_wait_poll(  );

unsigned char pio_inbyte( );
static void pio_outbyte(  );
static unsigned int pio_inword( );
static void pio_outword(  );

static void pio_drq_block_in( );
static void pio_drq_block_out(  );
static void pio_rep_inbyte( );
static void pio_rep_outbyte(  );
static void pio_rep_inword(  );
static void pio_rep_outword( );

static long tmr_cmd_start_time;     /* command start time */
static void tmr_set_timeout( );
static int tmr_chk_timeout( );

static int exec_non_data_cmd( );
static int exec_pio_data_in_cmd( );
static int exec_pio_data_out_cmd( );

/* This macro provides a small delay that is used in several */
/* places in the ATA command protocols: */
/*
#define DELAY400NS  { pio_inbyte( CB_ASTAT ); pio_inbyte( CB_ASTAT ); pio_inbyte( CB_ASTAT ); pio_inbyte( CB_ASTAT ); }
*/
/* Amiga shouldn't require a delay */
#define DELAY400NS	do {} while(0)

/************************************************************** */
/* */
/* reg_config() - Check the host adapter and determine the */
/*                number and type of drives attached. */
/* */
/* This process is not documented by any of the ATA standards. */
/* */
/* Infomation is returned by this function is in */
/* reg_config_info[] -- see MINDRVR.H. */
/* */
/************************************************************** */

int reg_config( )
{
   int numDev = 0;
   unsigned char sc;
   unsigned char sn;
   unsigned char cl;
   unsigned char ch;
   unsigned char st;
   unsigned char dc;

   /* setup register values */

   dc = (unsigned char) ( int_use_intr_flag ? 0 : CB_DC_NIEN );

   /* assume there are no devices */

   reg_config_info[0] = REG_CONFIG_TYPE_NONE;
   reg_config_info[1] = REG_CONFIG_TYPE_NONE;

   /* set up Device Control register */

   pio_outbyte( CB_DC, dc );

   /* lets see if there is a device 0 */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   pio_outbyte( CB_SC, 0xaa );
   pio_outbyte( CB_SN, 0x55 );
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x55 ) && ( sn == 0xaa ) )
      reg_config_info[0] = REG_CONFIG_TYPE_UNKN;

   /* lets see if there is a device 1 */

   pio_outbyte( CB_DH, CB_DH_DEV1 );
   DELAY400NS;
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   pio_outbyte( CB_SC, 0xaa );
   pio_outbyte( CB_SN, 0x55 );
   pio_outbyte( CB_SC, 0x55 );
   pio_outbyte( CB_SN, 0xaa );
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x55 ) && ( sn == 0xaa ) )
      reg_config_info[1] = REG_CONFIG_TYPE_UNKN;

   /* now we think we know which devices, if any are there, */
   /* so lets try a soft reset (ignoring any errors). */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   reg_reset( 0 );

   /* lets check device 0 again, is device 0 really there? */
   /* is it ATA or ATAPI? */

   pio_outbyte( CB_DH, CB_DH_DEV0 );
   DELAY400NS;
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
   {
      reg_config_info[0] = REG_CONFIG_TYPE_UNKN;
      st = pio_inbyte( CB_STAT );
      cl = pio_inbyte( CB_CL );
      ch = pio_inbyte( CB_CH );
      if ( ( ( cl == 0x14 ) && ( ch == 0xeb ) )       /* PATAPI */
           ||
           ( ( cl == 0x69 ) && ( ch == 0x96 ) )       /* SATAPI */
         )
      {
         reg_config_info[0] = REG_CONFIG_TYPE_ATAPI;
      }
      else
      if ( ( st != 0 )
           &&
           ( ( ( cl == 0x00 ) && ( ch == 0x00 ) )     /* PATA */
             ||
             ( ( cl == 0x3c ) && ( ch == 0xc3 ) ) )   /* SATA */
         )
      {
         reg_config_info[0] = REG_CONFIG_TYPE_ATA;
      }
   }

   /* lets check device 1 again, is device 1 really there? */
   /* is it ATA or ATAPI? */

   pio_outbyte( CB_DH, CB_DH_DEV1 );
   DELAY400NS;
   sc = pio_inbyte( CB_SC );
   sn = pio_inbyte( CB_SN );
   if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
   {
      reg_config_info[1] = REG_CONFIG_TYPE_UNKN;
      st = pio_inbyte( CB_STAT );
      cl = pio_inbyte( CB_CL );
      ch = pio_inbyte( CB_CH );
      if ( ( ( cl == 0x14 ) && ( ch == 0xeb ) )       /* PATAPI */
           ||
           ( ( cl == 0x69 ) && ( ch == 0x96 ) )       /* SATAPI */
         )
      {
         reg_config_info[1] = REG_CONFIG_TYPE_ATAPI;
      }
      else
      if ( ( st != 0 )
           &&
           ( ( ( cl == 0x00 ) && ( ch == 0x00 ) )     /* PATA */
             ||
             ( ( cl == 0x3c ) && ( ch == 0xc3 ) ) )   /* SATA */
         )
      {
         reg_config_info[1] = REG_CONFIG_TYPE_ATA;
      }
   }

   /* If possible, select a device that exists, try device 0 first. */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV1 );
      DELAY400NS;
      numDev ++ ;
   }
   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV0 );
      DELAY400NS;
      numDev ++ ;
   }

   /* return the number of devices found */

   return numDev;
}

/************************************************************** */
/* */
/* reg_reset() - Execute a Software Reset. */
/* */
/* See ATA-2 Section 9.2, ATA-3 Section 9.2, ATA-4 Section 8.3. */
/* */
/************************************************************** */

int reg_reset( devRtrn )
unsigned char devRtrn;
{
   unsigned char sc;
   unsigned char sn;
   unsigned char status;
   unsigned char dc;

   /* setup register values */

   dc = (unsigned char) ( int_use_intr_flag ? 0 : CB_DC_NIEN );

   /* initialize the command timeout counter */

   tmr_set_timeout();

   /* Set and then reset the soft reset bit in the Device Control */
   /* register.  This causes device 0 be selected. */

   pio_outbyte( CB_DC, (unsigned char) ( dc | CB_DC_SRST ) );
   DELAY400NS;
   pio_outbyte( CB_DC, dc );
   DELAY400NS;

   /* If there is a device 0, wait for device 0 to set BSY=0. */

   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      while ( 1 )
      {
         status = pio_inbyte( CB_STAT );
         if ( ( status & CB_STAT_BSY ) == 0 )
            break;
         if ( tmr_chk_timeout() )
         {
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 1;
            break;
         }
      }
   }

   /* If there is a device 1, wait until device 1 allows */
   /* register access. */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      while ( 1 )
      {
         pio_outbyte( CB_DH, CB_DH_DEV1 );
         DELAY400NS;
         sc = pio_inbyte( CB_SC );
         sn = pio_inbyte( CB_SN );
         if ( ( sc == 0x01 ) && ( sn == 0x01 ) )
            break;
         if ( tmr_chk_timeout() )
         {
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 2;
            break;
         }
      }

      /* Now check if drive 1 set BSY=0. */

      if ( reg_cmd_info.ec == 0 )
      {
         if ( pio_inbyte( CB_STAT ) & CB_STAT_BSY )
         {
            reg_cmd_info.ec = 3;
         }
      }
   }

   /* RESET_DONE: */

   /* We are done but now we must select the device the caller */
   /* requested. This will cause */
   /* the correct data to be returned in reg_cmd_info. */

   pio_outbyte( CB_DH, (unsigned char) ( devRtrn ? CB_DH_DEV1 : CB_DH_DEV0 ) );
   DELAY400NS;

   /* If possible, select a device that exists, */
   /* try device 0 first. */

   if ( reg_config_info[1] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV1 );
      DELAY400NS;
   }
   if ( reg_config_info[0] != REG_CONFIG_TYPE_NONE )
   {
      pio_outbyte( CB_DH, CB_DH_DEV0 );
      DELAY400NS;
   }

   /* All done.  The return values of this function are described in */
   /* MINDRVR.H. */

   sub_trace_command();
   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/************************************************************** */
/* */
/* exec_non_data_cmd() - Execute a non-data command. */
/* */
/* This includes the strange ATAPI DEVICE RESET 'command' */
/* (command code 08H). */
/* */
/* Note special handling for Execute Device Diagnostics */
/* command when there is no device 0. */
/* */
/* See ATA-2 Section 9.5, ATA-3 Section 9.5, */
/* ATA-4 Section 8.8 Figure 12.  Also see Section 8.5. */
/* */
/************************************************************** */



static int exec_non_data_cmd( dev )
unsigned char dev;
{
   unsigned char secCnt;
   unsigned char secNum;
   unsigned char status;
   int polled = 0;

   /* Set command time out. */

   tmr_set_timeout();

   /* PAY ATTENTION HERE */
   /* If the caller is attempting a Device Reset command, then */
   /* don't do most of the normal stuff.  Device Reset has no */
   /* parameters, should not generate an interrupt and it is the */
   /* only command that can be written to the Command register */
   /* when a device has BSY=1 (a very strange command!).  Not */
   /* all devices support this command (even some ATAPI devices */
   /* don't support the command. */

   if ( reg_cmd_info.cmd != CMD_DEVICE_RESET )
   {
      /* Select the drive - call the sub_select function. */
      /* Quit now if this fails. */

      if ( sub_select( dev ) )
      {
         return 1;
      }

      /* Set up all the registers except the command register. */

      sub_setup_command();
   }

   /* Start the command by setting the Command register.  The drive */
   /* should immediately set BUSY status. */

   pio_outbyte( CB_CMD, reg_cmd_info.cmd );

   /* Waste some time by reading the alternate status a few times. */
   /* This gives the drive time to set BUSY in the status register on */
   /* really fast systems.  If we don't do this, a slow drive on a fast */
   /* system may not set BUSY fast enough and we would think it had */
   /* completed the command when it really had not even started the */
   /* command yet. */

   DELAY400NS;

   /* IF */
   /*    This is an Exec Dev Diag command (cmd=0x90) */
   /*    and there is no device 0 then */
   /*    there will be no interrupt. So we must */
   /*    poll device 1 until it allows register */
   /*    access and then do normal polling of the Status */
   /*    register for BSY=0. */
   /* ELSE */
   /* IF */
   /*    This is a Dev Reset command (cmd=0x08) then */
   /*    there should be no interrupt.  So we must */
   /*    poll for BSY=0. */
   /* ELSE */
   /*    Do the normal wait for interrupt or polling for */
   /*    completion. */

   if ( ( reg_cmd_info.cmd == CMD_EXECUTE_DEVICE_DIAGNOSTIC )
        &&
        ( reg_config_info[0] == REG_CONFIG_TYPE_NONE )
      )
   {
      polled = 1;
      while ( 1 )
      {
         pio_outbyte( CB_DH, CB_DH_DEV1 );
         DELAY400NS;
         secCnt = pio_inbyte( CB_SC );
         secNum = pio_inbyte( CB_SN );
         if ( ( secCnt == 0x01 ) && ( secNum == 0x01 ) )
            break;
         if ( tmr_chk_timeout() )
         {
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = 24;
            break;
         }
      }
   }
   else
   {
      if ( reg_cmd_info.cmd == CMD_DEVICE_RESET )
      {
         /* Wait for not BUSY -or- wait for time out. */

         polled = 1;
         sub_wait_poll( 0, 23 );
      }
      else
      {
         /* Wait for interrupt -or- wait for not BUSY -or- wait for time out. */

         if ( ! int_use_intr_flag )
            polled = 1;
         sub_wait_poll( 22, 23 );
      }
   }

   /* If status was polled or if any error read the status register, */
   /* otherwise get the status that was read by the interrupt handler. */

   if ( ( polled ) || ( reg_cmd_info.ec ) )
      status = pio_inbyte( CB_STAT );
   else
      status = int_ata_status;

   /* Error if BUSY, DEVICE FAULT, DRQ or ERROR status now. */

   if ( reg_cmd_info.ec == 0 )
   {
      if ( status & ( CB_STAT_BSY | CB_STAT_DF | CB_STAT_DRQ | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 21;
      }
   }

   /* NON_DATA_DONE: */

   /* All done.  The return values of this function are described in */
   /* MINDRVR.H. */

   sub_trace_command();
   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/************************************************************** */
/* */
/* reg_non_data_lba28() - Easy way to execute a non-data command */
/*                        using an LBA sector address. */
/* */
/************************************************************** */

int reg_non_data_lba28( dev, cmd, fr, sc, lba )
unsigned char dev; 
unsigned char cmd;                        
unsigned int fr; 
unsigned int sc;
unsigned long lba; 
{

   /* Setup current command information. */

   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr = fr;
   reg_cmd_info.sc = sc;
   reg_cmd_info.dh = (unsigned char) ( CB_DH_LBA | ( dev ? CB_DH_DEV1 : CB_DH_DEV0 ) );
   reg_cmd_info.dc = (unsigned char) ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   reg_cmd_info.ns  = sc;
   reg_cmd_info.lbaSize = LBA28;
   reg_cmd_info.lbaHigh = 0L;
   reg_cmd_info.lbaLow = lba;

   /* Execute the command. */

   return exec_non_data_cmd( dev );
}


/************************************************************** */
/* */
/* exec_pio_data_in_cmd() - Execute a PIO Data In command. */
/* */
/* See ATA-2 Section 9.3, ATA-3 Section 9.3, */
/* ATA-4 Section 8.6 Figure 10. */
/* */
/************************************************************** */

static int exec_pio_data_in_cmd( dev, bufAddr, numSect, multiCnt )
unsigned char dev;
unsigned char * bufAddr;
long numSect; 
int multiCnt; 
{
   unsigned char status;
   long wordCnt;

   /* Set command time out. */

   tmr_set_timeout();

   /* Select the drive - call the sub_select function. */
   /* Quit now if this fails. */

   if ( sub_select( dev ) )
   {
      return 1;
   }

   /* Set up all the registers except the command register. */

   sub_setup_command();

   /* Start the command by setting the Command register.  The drive */
   /* should immediately set BUSY status. */

   pio_outbyte( CB_CMD, reg_cmd_info.cmd );

   /* Waste some time by reading the alternate status a few times. */
   /* This gives the drive time to set BUSY in the status register on */
   /* really fast systems.  If we don't do this, a slow drive on a fast */
   /* system may not set BUSY fast enough and we would think it had */
   /* completed the command when it really had not even started the */
   /* command yet. */

   DELAY400NS;

   /* Loop to read each sector. */

   while ( 1 )
   {
      /* READ_LOOP: */
      /* */
      /* NOTE NOTE NOTE ...  The primary status register (1f7) MUST NOT be */
      /* read more than ONCE for each sector transferred!  When the */
      /* primary status register is read, the drive resets IRQ.  The */
      /* alternate status register (3f6) can be read any number of times. */
      /* After interrupt read the the primary status register ONCE */
      /* and transfer the 256 words (REP INSW).  AS SOON as BOTH the */
      /* primary status register has been read AND the last of the 256 */
      /* words has been read, the drive is allowed to generate the next */
      /* IRQ (newer and faster drives could generate the next IRQ in */
      /* 50 microseconds or less).  If the primary status register is read */
      /* more than once, there is the possibility of a race between the */
      /* drive and the software and the next IRQ could be reset before */
      /* the system interrupt controller sees it. */

      /* Wait for interrupt -or- wait for not BUSY -or- wait for time out. */

      sub_wait_poll( 34, 35 );

      /* If polling or error read the status, otherwise */
      /* get the status that was read by the interrupt handler. */

      if ( ( ! int_use_intr_flag ) || ( reg_cmd_info.ec ) )
         status = pio_inbyte( CB_STAT );
      else
         status = int_ata_status;

      /* If there was a time out error, go to READ_DONE. */

      if ( reg_cmd_info.ec )
         break;   /* go to READ_DONE */

      /* If BSY=0 and DRQ=1, transfer the data, */
      /* even if we find out there is an error later. */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == CB_STAT_DRQ )
      {

         /* increment number of DRQ packets */

         reg_cmd_info.drqPackets ++ ;

         /* determine the number of sectors to transfer */

         wordCnt = multiCnt ? multiCnt : 1;
         if ( wordCnt > numSect )
            wordCnt = numSect;
         wordCnt = wordCnt * 256;

         /* Do the REP INSW to read the data for one DRQ block. */

         reg_cmd_info.totalBytesXfer += ( wordCnt << 1 );
         pio_drq_block_in( CB_DATA, bufAddr, wordCnt );

         DELAY400NS;    /* delay so device can get the status updated */

         /* Note: The drive should have dropped DATA REQUEST by now.  If there */
         /* are more sectors to transfer, BUSY should be active now (unless */
         /* there is an error). */

         /* Decrement the count of sectors to be transferred */
         /* and increment buffer address. */

         numSect = numSect - ( multiCnt ? multiCnt : 1 );
         bufAddr = bufAddr + ( 512 * ( multiCnt ? multiCnt : 1 ) );
      }

      /* So was there any error condition? */

      if ( status & ( CB_STAT_BSY | CB_STAT_DF | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 31;
         break;   /* go to READ_DONE */
      }

      /* DRQ should have been set -- was it? */

      if ( ( status & CB_STAT_DRQ ) == 0 )
      {
         reg_cmd_info.ec = 32;
         break;   /* go to READ_DONE */
      }

      /* If all of the requested sectors have been transferred, make a */
      /* few more checks before we exit. */

      if ( numSect < 1 )
      {
         /* Since the drive has transferred all of the requested sectors */
         /* without error, the drive should not have BUSY, DEVICE FAULT, */
         /* DATA REQUEST or ERROR active now. */

         status = pio_inbyte( CB_STAT );
         if ( status & ( CB_STAT_BSY | CB_STAT_DF | CB_STAT_DRQ | CB_STAT_ERR ) )
         {
            reg_cmd_info.ec = 33;
            break;   /* go to READ_DONE */
         }

         /* All sectors have been read without error, go to READ_DONE. */

         break;   /* go to READ_DONE */

      }

      /* This is the end of the read loop.  If we get here, the loop is */
      /* repeated to read the next sector.  Go back to READ_LOOP. */

   }

   /* READ_DONE: */

   /* All done.  The return values of this function are described in */
   /* MINDRVR.H. */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/************************************************************** */
/* */
/* reg_pio_data_in_lba28() - Easy way to execute a PIO Data In command */
/*                           using an LBA sector address. */
/* */
/************************************************************** */

int reg_pio_data_in_lba28( dev, cmd, fr, sc, lba, bufAddr, numSect, multiCnt )
unsigned char dev; 
unsigned char cmd;
unsigned int fr; 
unsigned int sc;
unsigned long lba;
unsigned char * bufAddr;
long numSect; 
int multiCnt;
{

   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr = fr;
   reg_cmd_info.sc = sc;
   reg_cmd_info.dh = (unsigned char) ( CB_DH_LBA | ( dev ? CB_DH_DEV1 : CB_DH_DEV0 ) );
   reg_cmd_info.dc = (unsigned char) ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   reg_cmd_info.lbaSize = LBA28;
   reg_cmd_info.lbaHigh = 0L;
   reg_cmd_info.lbaLow = lba;

   /* these commands transfer only 1 sector */
   if (    ( cmd == CMD_IDENTIFY_DEVICE )
        || ( cmd == CMD_IDENTIFY_DEVICE_PACKET )
      )
      numSect = 1;

   /* adjust multiple count */
   if ( multiCnt & 0x0800 )
   {
      /* assume caller knows what they are doing */
      multiCnt &= 0x00ff;
   }
   else
   {
      /* only Read Multiple uses multiCnt */
      if ( cmd != CMD_READ_MULTIPLE )
         multiCnt = 1;
   }

   reg_cmd_info.ns  = numSect;
   reg_cmd_info.mc  = multiCnt;

   return exec_pio_data_in_cmd( dev, bufAddr, numSect, multiCnt );
}


/************************************************************** */
/* */
/* exec_pio_data_out_cmd() - Execute a PIO Data Out command. */
/* */
/* See ATA-2 Section 9.4, ATA-3 Section 9.4, */
/* ATA-4 Section 8.7 Figure 11. */
/* */
/************************************************************** */



static int exec_pio_data_out_cmd( dev, bufAddr, numSect, multiCnt )
unsigned char dev;
unsigned char * bufAddr;
long numSect; 
int multiCnt; 
{
   unsigned char status;
   int loopFlag = 1;
   long wordCnt;

   /* Set command time out. */

   tmr_set_timeout();

   /* Select the drive - call the sub_select function. */
   /* Quit now if this fails. */

   if ( sub_select( dev ) )
   {
      return 1;
   }

   /* Set up all the registers except the command register. */

   sub_setup_command();

   /* Start the command by setting the Command register.  The drive */
   /* should immediately set BUSY status. */

   pio_outbyte( CB_CMD, reg_cmd_info.cmd );

   /* Waste some time by reading the alternate status a few times. */
   /* This gives the drive time to set BUSY in the status register on */
   /* really fast systems.  If we don't do this, a slow drive on a fast */
   /* system may not set BUSY fast enough and we would think it had */
   /* completed the command when it really had not even started the */
   /* command yet. */

   DELAY400NS;

   /* Wait for not BUSY or time out. */
   /* Note: No interrupt is generated for the */
   /* first sector of a write command. */

   while ( 1 )
   {
      status = pio_inbyte( CB_ASTAT );
      if ( ( status & CB_STAT_BSY ) == 0 )
         break;
      if ( tmr_chk_timeout() )
      {
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 47;
         loopFlag = 0;
         break;
      }
   }

   /* This loop writes each sector. */

   while ( loopFlag )
   {
      /* WRITE_LOOP: */
      /* */
      /* NOTE NOTE NOTE ...  The primary status register (1f7) MUST NOT be */
      /* read more than ONCE for each sector transferred!  When the */
      /* primary status register is read, the drive resets IRQ.  The */
      /* alternate status register (3f6) can be read any number of times. */
      /* For correct results, transfer the 256 words (REP OUTSW), wait for */
      /* interrupt and then read the primary status register.  AS */
      /* SOON as BOTH the primary status register has been read AND the */
      /* last of the 256 words has been written, the drive is allowed to */
      /* generate the next IRQ (newer and faster drives could generate */
      /* the next IRQ in 50 microseconds or less).  If the primary */
      /* status register is read more than once, there is the possibility */
      /* of a race between the drive and the software and the next IRQ */
      /* could be reset before the system interrupt controller sees it. */

      /* If BSY=0 and DRQ=1, transfer the data, */
      /* even if we find out there is an error later. */

      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == CB_STAT_DRQ )
      {

         /* increment number of DRQ packets */

         reg_cmd_info.drqPackets ++ ;

         /* determine the number of sectors to transfer */

         wordCnt = multiCnt ? multiCnt : 1;
         if ( wordCnt > numSect )
            wordCnt = numSect;
         wordCnt = wordCnt * 256;

         /* Do the REP OUTSW to write the data for one DRQ block. */

         reg_cmd_info.totalBytesXfer += ( wordCnt << 1 );
         pio_drq_block_out( CB_DATA, bufAddr, wordCnt );

         DELAY400NS;    /* delay so device can get the status updated */

         /* Note: The drive should have dropped DATA REQUEST and */
         /* raised BUSY by now. */

         /* Decrement the count of sectors to be transferred */
         /* and increment buffer address. */

         numSect = numSect - ( multiCnt ? multiCnt : 1 );
         bufAddr = bufAddr + ( 512 * ( multiCnt ? multiCnt : 1 ) );
      }

      /* So was there any error condition? */

      if ( status & ( CB_STAT_BSY | CB_STAT_DF | CB_STAT_ERR ) )
      {
         reg_cmd_info.ec = 41;
         break;   /* go to WRITE_DONE */
      }

      /* DRQ should have been set -- was it? */

      if ( ( status & CB_STAT_DRQ ) == 0 )
      {
         reg_cmd_info.ec = 42;
         break;   /* go to WRITE_DONE */
      }

      /* Wait for interrupt -or- wait for not BUSY -or- wait for time out. */

      sub_wait_poll( 44, 45 );

      /* If polling or error read the status, otherwise */
      /* get the status that was read by the interrupt handler. */

      if ( ( ! int_use_intr_flag ) || ( reg_cmd_info.ec ) )
         status = pio_inbyte( CB_STAT );
      else
         status = int_ata_status;

      /* If there was a time out error, go to WRITE_DONE. */

      if ( reg_cmd_info.ec )
         break;   /* go to WRITE_DONE */

      /* If all of the requested sectors have been transferred, make a */
      /* few more checks before we exit. */

      if ( numSect < 1 )
      {
         /* Since the drive has transferred all of the sectors without */
         /* error, the drive MUST not have BUSY, DEVICE FAULT, DATA REQUEST */
         /* or ERROR status at this time. */

         if ( status & ( CB_STAT_BSY | CB_STAT_DF | CB_STAT_DRQ | CB_STAT_ERR ) )
         {
            reg_cmd_info.ec = 43;
            break;   /* go to WRITE_DONE */
         }

         /* All sectors have been written without error, go to WRITE_DONE. */

         break;   /* go to WRITE_DONE */

      }

      /* */
      /* This is the end of the write loop.  If we get here, the loop */
      /* is repeated to write the next sector.  Go back to WRITE_LOOP. */

   }

   /* WRITE_DONE: */

   /* All done.  The return values of this function are described in */
   /* MINDRVR.H. */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/************************************************************** */
/* */
/* reg_pio_data_out_lba28() - Easy way to execute a PIO Data In command */
/*                            using an LBA sector address. */
/* */
/************************************************************** */

int reg_pio_data_out_lba28( dev, cmd, fr, sc, lba, bufAddr, numSect, multiCnt )
unsigned char dev; 
unsigned char cmd;
unsigned int fr; 
unsigned int sc;
unsigned long lba;
unsigned char * bufAddr;
long numSect; 
int multiCnt; 
{

   reg_cmd_info.cmd = cmd;
   reg_cmd_info.fr = fr;
   reg_cmd_info.sc = sc;
   reg_cmd_info.dh = (unsigned char) ( CB_DH_LBA | ( dev ? CB_DH_DEV1 : CB_DH_DEV0 ) );
   reg_cmd_info.dc = (unsigned char) ( int_use_intr_flag ? 0 : CB_DC_NIEN );
   reg_cmd_info.lbaSize = LBA28;
   reg_cmd_info.lbaHigh = 0;
   reg_cmd_info.lbaLow = lba;

   /* adjust multiple count */
   if ( multiCnt & 0x0800 )
   {
      /* assume caller knows what they are doing */
      multiCnt &= 0x00ff;
   }
   else
   {
      /* only Write Multiple and CFA Write Multiple W/O Erase uses multiCnt */
      if (    ( cmd != CMD_WRITE_MULTIPLE )
           && ( cmd != CMD_CFA_WRITE_MULTIPLE_WO_ERASE )
         )
         multiCnt = 1;
   }

   reg_cmd_info.ns  = numSect;
   reg_cmd_info.mc  = multiCnt;

   return exec_pio_data_out_cmd( dev, bufAddr, numSect, multiCnt );
}


/************************************************************** */
/* */
/* sub_setup_command() -- setup the command parameters */
/*                        in FR, SC, SN, CL, CH and DH. */
/* */
/************************************************************** */

static void sub_setup_command( )
{

   /* output DevCtrl - same for all devices and commands */
   pio_outbyte( CB_DC, reg_cmd_info.dc );

   /* output command parameters */

      /* in ATA LBA28 mode */
   pio_outbyte( CB_FR, (unsigned char) reg_cmd_info.fr );
   pio_outbyte( CB_SC, (unsigned char) reg_cmd_info.sc );
   pio_outbyte( CB_SN, (unsigned char) reg_cmd_info.lbaLow );
   pio_outbyte( CB_CL, (unsigned char) ( reg_cmd_info.lbaLow >> 8 ) );
   pio_outbyte( CB_CH, (unsigned char) ( reg_cmd_info.lbaLow >> 16 ) );
   pio_outbyte( CB_DH, (unsigned char) ( ( reg_cmd_info.dh & 0xf0 )
                                            | ( ( reg_cmd_info.lbaLow >> 24 ) & 0x0f ) ) );


}

/************************************************************** */
/* */
/* sub_trace_command() -- trace the end of a command. */
/* */
/************************************************************** */

static void sub_trace_command( )
{

   reg_cmd_info.st = pio_inbyte( CB_STAT );
   reg_cmd_info.as = pio_inbyte( CB_ASTAT );
   reg_cmd_info.er = pio_inbyte( CB_ERR );

/* !!! if you want to read back the other device registers */
/* !!! at the end of a command then this is the place to do */
/* !!! it. The code here is just and example of out this is */
/* !!! done on a little endian system like an x86. */

#if 0    /* read back other registers */

   {
      unsigned long lbaHigh;
      unsigned long lbaLow;

      lbaHigh = 0;
      lbaLow = 0;

         /* read back ATA LBA28 */
      lbaLow = pio_inbyte( CB_DH );
      lbaLow = lbaLow << 8;
      lbaLow = lbaLow | pio_inbyte( CB_CH );
      lbaLow = lbaLow << 8;
      lbaLow = lbaLow | pio_inbyte( CB_CL );
      lbaLow = lbaLow << 8;
      lbaLow = lbaLow | pio_inbyte( CB_SN );

   }

#endif   /* read back other registers */

}

/************************************************************** */
/* */
/* sub_select() - function used to select a drive. */
/* */
/* Function to select a drive making sure that BSY=0 and DRQ=0. */
/* */
/*************************************************************** */

static int sub_select( dev )
unsigned char dev;
{
   unsigned char status;

   /* PAY ATTENTION HERE */
   /* The caller may want to issue a command to a device that doesn't */
   /* exist (for example, Exec Dev Diag), so if we see this, */
   /* just select that device, skip all status checking and return. */
   /* We assume the caller knows what they are doing! */

   if ( reg_config_info[dev] < REG_CONFIG_TYPE_ATA )
   {
      /* select the device and return */

      pio_outbyte( CB_DH, (unsigned char) ( dev ? CB_DH_DEV1 : CB_DH_DEV0 ) );
      DELAY400NS;
      return 0;
   }

   /* The rest of this is the normal ATA stuff for device selection */
   /* and we don't expect the caller to be selecting a device that */
   /* does not exist. */
   /* We don't know which drive is currently selected but we should */
   /* wait BSY=0 and DRQ=0. Normally both BSY=0 and DRQ=0 */
   /* unless something is very wrong! */

   while ( 1 )
   {
      status = pio_inbyte( CB_STAT );
      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == 0 )
         break;
      if ( tmr_chk_timeout() )
      {
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 11;
         reg_cmd_info.st = status;
         reg_cmd_info.as = pio_inbyte( CB_ASTAT );
         reg_cmd_info.er = pio_inbyte( CB_ERR );
         return 1;
      }
   }

   /* Here we select the drive we really want to work with by */
   /* setting the DEV bit in the Drive/Head register. */

   pio_outbyte( CB_DH, (unsigned char) ( dev ? CB_DH_DEV1 : CB_DH_DEV0 ) );
   DELAY400NS;

   /* Wait for the selected device to have BSY=0 and DRQ=0. */
   /* Normally the drive should be in this state unless */
   /* something is very wrong (or initial power up is still in */
   /* progress). */

   while ( 1 )
   {
      status = pio_inbyte( CB_STAT );
      if ( ( status & ( CB_STAT_BSY | CB_STAT_DRQ ) ) == 0 )
         break;
      if ( tmr_chk_timeout() )
      {
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = 12;
         reg_cmd_info.st = status;
         reg_cmd_info.as = pio_inbyte( CB_ASTAT );
         reg_cmd_info.er = pio_inbyte( CB_ERR );
         return 1;
      }
   }

   /* All done.  The return values of this function are described in */
   /* ATAIO.H. */

   if ( reg_cmd_info.ec )
      return 1;
   return 0;
}

/************************************************************** */
/* */
/* sub_wait_poll() - wait for interrupt or poll for BSY=0 */
/* */
/************************************************************** */

static void sub_wait_poll( we, pe )
unsigned char we; 
unsigned char pe;
{
   unsigned char status;

   /* Wait for interrupt -or- wait for not BUSY -or- wait for time out. */

   if ( we && int_use_intr_flag )
   {
      if ( SYSTEM_WAIT_INTR_OR_TIMEOUT() )    /* time out ? */
      {
         reg_cmd_info.to = 1;
         reg_cmd_info.ec = we;
      }
   }
   else
   {
      while ( 1 )
      {
         status = pio_inbyte( CB_ASTAT );       /* poll for not busy */
         if ( ( status & CB_STAT_BSY ) == 0 )
            break;
         if ( tmr_chk_timeout() )               /* time out yet ? */
         {
            reg_cmd_info.to = 1;
            reg_cmd_info.ec = pe;
            break;
         }
      }
   }
}

/************************************************************** */
/* */
/* These functions do basic IN/OUT of byte and word values: */
/* */
/*    pio_inbyte() */
/*    pio_outbyte() */
/*    pio_inword() */
/*    pio_outword() */
/* */
/************************************************************** */

unsigned char pio_inbyte( addr )
unsigned char addr;
{
   /*!!! read an 8-bit ATA register */

   return * pio_reg_addrs[ addr ];
}

/************************************************************** */

static void pio_outbyte( addr, data )
int addr; 
unsigned char data;
{

   /*!!! write an 8-bit ATA register */

   * pio_reg_addrs[ addr ] = data;
}

/************************************************************** */

static unsigned int pio_inword( addr )
unsigned char addr;
{

   /*!!! read a 16-bit ATA register (usually the ATA Data register) */

   return * ( (unsigned int *) pio_reg_addrs[ addr ] );
}

/************************************************************** */

static void pio_outword( addr, data )
int addr; 
unsigned int data;
{

   /*!!! Write a 16-bit ATA register (usually the ATA Data register) */

   * ( (unsigned int *) pio_reg_addrs[ addr ] ) = data;
}


/************************************************************** */
/* */
/* These functions are normally used to transfer DRQ blocks: */
/* */
/* pio_drq_block_in() */
/* pio_drq_block_out() */
/* */
/************************************************************** */

/* Note: pio_drq_block_in() is the primary way perform PIO */
/* Data In transfers. It will handle 8-bit, 16-bit and 32-bit */
/* I/O based data transfers. */

static void pio_drq_block_in( addrDataReg, bufAddr, wordCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long wordCnt;
{

   /* NOTE: wordCnt is the size of a DRQ data block/packet */
   /* in words. The maximum value of wordCnt is normally: */
   /* a) For ATA, 16384 words or 32768 bytes (64 sectors, */
   /*    only with READ/WRITE MULTIPLE commands), */
   /* b) For ATAPI, 32768 words or 65536 bytes */
   /*    (actually 65535 bytes plus a pad byte). */

   {
      int pxw;
      long wc;

      /* adjust pio_xfer_width - don't use DWORD if wordCnt is odd. */

      pxw = pio_xfer_width;

      /* Data transfer using INS instruction. */
      /* Break the transfer into chunks of 32768 or fewer bytes. */

      while ( wordCnt > 0 )
      {
         if ( wordCnt > 16384L )
            wc = 16384;
         else
            wc = wordCnt;
         if ( pxw == 8 )
         {
            /* do REP INS */
            pio_rep_inbyte( addrDataReg, bufAddr, wc * 2L );
         }
         else
         {
            /* do REP INSW */
            pio_rep_inword( addrDataReg, bufAddr, wc );
         }
         bufAddr = bufAddr + ( wc * 2 );
         wordCnt = wordCnt - wc;
      }
   }

   return;
}

/************************************************************** */

/* Note: pio_drq_block_out() is the primary way perform PIO */
/* Data Out transfers. It will handle 8-bit, 16-bit and 32-bit */
/* I/O based data transfers. */

static void pio_drq_block_out( addrDataReg, bufAddr, wordCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long wordCnt;
{

   /* NOTE: wordCnt is the size of a DRQ data block/packet */
   /* in words. The maximum value of wordCnt is normally: */
   /* a) For ATA, 16384 words or 32768 bytes (64 sectors, */
   /*    only with READ/WRITE MULTIPLE commands), */
   /* b) For ATAPI, 32768 words or 65536 bytes */
   /*    (actually 65535 bytes plus a pad byte). */

   {
      int pxw;
      long wc;

      /* adjust pio_xfer_width - don't use DWORD if wordCnt is odd. */

      pxw = pio_xfer_width;

      /* Data transfer using OUTS instruction. */
      /* Break the transfer into chunks of 32768 or fewer bytes. */

      while ( wordCnt > 0 )
      {
         if ( wordCnt > 16384L )
            wc = 16384;
         else
            wc = wordCnt;
         if ( pxw == 8 )
         {
            /* do REP OUTS */
            pio_rep_outbyte( addrDataReg, bufAddr, wc * 2L );
         }
         else
         {
            /* do REP OUTSW */
            pio_rep_outword( addrDataReg, bufAddr, wc );
         }
         bufAddr = bufAddr + ( wc * 2 );
         wordCnt = wordCnt - wc;
      }
   }

   return;
}

/************************************************************** */
/* */
/* These functions transfer PIO DRQ data blocks through the ATA */
/* Data register. On an x86 these functions would use the */
/* REP INS and REP OUTS instructions. */
/* */
/* pio_rep_inbyte() */
/* pio_rep_outbyte() */
/* pio_rep_inword() */
/* pio_rep_outword() */
/* */
/* These functions can be called directly but usually they */
/* are called by the pio_drq_block_in() and pio_drq_block_out() */
/* functions to perform I/O mode transfers. See the */
/* pio_xfer_width variable! */
/* */
/************************************************************** */

static void pio_rep_inbyte( addrDataReg, bufAddr, byteCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long byteCnt;
{

   /* Warning: Avoid calling this function with */
   /* byteCnt > 32768 (transfers 32768 bytes). */
   /* that bufOff is a value between 0 and 15 (0xf). */

   /*!!! repeat read an 8-bit register (ATA Data register when */
   /*!!! ATA status is BSY=0 DRQ=1). For example: */

   while ( byteCnt > 0 )
   {
      * bufAddr = pio_inbyte( addrDataReg );
      bufAddr ++ ;
      byteCnt -- ;
   }
}

/************************************************************** */

static void pio_rep_outbyte( addrDataReg, bufAddr, byteCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long byteCnt;
{

   /* Warning: Avoid calling this function with */
   /* byteCnt > 32768 (transfers 32768 bytes). */
   /* that bufOff is a value between 0 and 15 (0xf). */

   /*!!! repeat write an 8-bit register (ATA Data register when */
   /*!!! ATA status is BSY=0 DRQ=1). For example: */

   while ( byteCnt > 0 )
   {
      pio_outbyte( addrDataReg, * bufAddr );
      bufAddr ++ ;
      byteCnt -- ;
   }
}

/************************************************************** */

static void pio_rep_inword( addrDataReg, bufAddr, wordCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long wordCnt;
{

   /* Warning: Avoid calling this function with */
   /* wordCnt > 16384 (transfers 32768 bytes). */

   /*!!! repeat read a 16-bit register (ATA Data register when */
   /*!!! ATA status is BSY=0 DRQ=1). For example: */

   while ( wordCnt > 0 )
   {
      * (unsigned int *) bufAddr = pio_inword( addrDataReg );
      bufAddr += 2;
      wordCnt -- ;
   }
}

/************************************************************** */

static void pio_rep_outword( addrDataReg, bufAddr, wordCnt )
unsigned char addrDataReg;
unsigned char * bufAddr;
long wordCnt;
{

   /* Warning: Avoid calling this function with */
   /* wordCnt > 16384 (transfers 32768 bytes). */

   /*!!! repeat write a 16-bit register (ATA Data register when */
   /*!!! ATA status is BSY=0 DRQ=1). For example: */

   while ( wordCnt > 0 )
   {
      pio_outword( addrDataReg, * (unsigned int *) bufAddr );
      bufAddr += 2;
      wordCnt -- ;
   }
}


/************************************************************** */
/* */
/* Command timing functions */
/* */
/*************************************************************** */

/************************************************************** */
/* */
/* tmr_set_timeout() - get the command start time */
/* */
/*************************************************************** */

static void tmr_set_timeout( void )

{

   /* get the command start time */
   tmr_cmd_start_time = SYSTEM_READ_TIMER();
}

/************************************************************** */
/* */
/* tmr_chk_timeout() - check for command timeout. */
/* */
/* Gives non-zero return if command has timed out. */
/* */
/*************************************************************** */

static int tmr_chk_timeout( void )

{
   long curTime;

   /* get current time */
   curTime = SYSTEM_READ_TIMER();

   /* timed out yet ? */
   if ( curTime >= ( tmr_cmd_start_time
                     + ( TMR_TIME_OUT * SYSTEM_TIMER_TICKS_PER_SECOND ) ) )
      return 1;      /* yes */

   /* no timeout yet */
   return 0;
}



/* end mindrvr.c */
